# SPDX-License-Identifier: MIT
from ...utils import chexdump32
from construct import *
import struct
import time
from .isp_opcodes import *

class ISPIORequestCommand:
    def __init__(self, iova, insize, outsize, args):
        self.iova = iova
        self.insize = insize
        self.outsize = outsize
        self.args = args
        self.opcode = struct.unpack("<l", self.args[0x4:0x8])[0]

class ISPIOCommandDispatcher:
    def __init__(self, isp):
        self.isp = isp
        self.opcode_dict = opcode_dict
        self.cmd_iova = self.isp.cmd_iova
        self._stfu = False

    @property
    def stfu(self): self._stfu = True

    def log(self, *args):
        if (not self._stfu):
            if (args): print("ISP:", *args)
            else: print()

    def send(self, cmd, cb=None):
        # TX: REQ: [0x1813140, 0xc, 0xc]  # iova, insize, outsize
        # RX: RSP: [0x1813141, 0xc, 0x0]  # iova, insize, zero
        req = ISPChannelMessage.build(
            arg0 = cmd.iova, # iova
            arg1 = cmd.insize,  # insize
            arg2 = cmd.outsize,  # outsize
        )
        # print("CHAN: IO: TX: REQ: [iova: 0x%x, insize: 0x%x, outsize: 0x%x]" % (req.arg0, req.arg1, req.arg2))

        # The largest (used) struct is 0x118. This should be good.
        patch = struct.pack("<l", 0x0)*(0x200//4)
        self.isp.iowrite(req.arg0, patch)
        self.isp.iowrite(req.arg0, cmd.args)

        rsp = self.isp.table.io.send(req)
        if (rsp == None):
            self.isp.table.terminal.dump()
            self.isp.table.dump()
            raise RuntimeError("Command %s [0x%04x] failed!" % (self.opcode2name(cmd.opcode), cmd.opcode))
        else:
            self.log("Command %s [0x%04x] success!" % (self.opcode2name(cmd.opcode), cmd.opcode))
            if (cb): cb()
        return rsp

    def opcode2name(self, opcode):
        if opcode in self.opcode_dict: return self.opcode_dict[opcode]
        else: return "CISP_CMD_UNKNOWN_%04x" % (opcode)

    def cmd_print_enable(self):
        s_cmd_print_enable = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "enable" / Int32ul,
        )
        args = s_cmd_print_enable.build(dict(
                opcode=CISP_CMD_PRINT_ENABLE,
                enable=0x1,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0xc,
            outsize=0xc,
            args=args,
        )
        return self.send(cmd)

    def cmd_trace_enable(self):
        s_cmd_trace_enable = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "arg1" / Int32ul,
                "arg2" / Int32ul,
        )
        args = s_cmd_trace_enable.build(dict(
                opcode=CISP_CMD_TRACE_ENABLE,
                arg1=0x0,
                arg2=0x0,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0xc,
            outsize=0xc,
            args=args,
        )
        return self.send(cmd)

    def cmd_set_isp_pmu_base(self):
        s_cmd_set_isp_pmu_base = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "base" / Int64ul,
        )
        args = s_cmd_set_isp_pmu_base.build(dict(
                opcode=CISP_CMD_SET_ISP_PMU_BASE,
                base=0x23b704000,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x10,
            outsize=0x10,
            args=args,
        )
        return self.send(cmd)

    def cmd_set_dsid_clr_req_base2(self):
        s_cmd_set_dsid_clr_req_base2 = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "base0" / Int64ul,
            "base1" / Int64ul,
            "base2" / Int64ul,
            "base3" / Int64ul,
            "regRange0" / Int32ul,
            "regRange1" / Int32ul,
            "regRange2" / Int32ul,
            "regRange3" / Int32ul,
        )
        args = s_cmd_set_dsid_clr_req_base2.build(dict(
                opcode=CISP_CMD_SET_DSID_CLR_REG_BASE2,
                base0=0x200014000,
            base1=0x200054000,
            base2=0x200094000,
            base3=0x2000d4000,
            regRange0=0x1000,
            regRange1=0x1000,
            regRange2=0x1000,
            regRange3=0x1000,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x38,
            outsize=0x0,
            args=args,
        )
        return self.send(cmd)

    def cmd_pmp_ctrl_set(self):
        """
        ISPCPU: PMP clock[scratch 0x23b738010,Int 0x23bc3c000(bit 1),size 4],bandwidth[scratch 0x23b73800c,Int 0x23bc3c000(bit 0),size 4
        Requested Vmin (300MHz) to PMP
        00000000  00000000 0000001c 3b738010 00000002 3bc3c000 00000002 00000401 3b73800c
        00000020  00000002 3bc3c000 00000002 00000400 00001000 00001000
        """
        s_cmd_pmp_ctrl_set = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "addr0" / Int64ul,
            "addr1" / Int64ul,
            "unk_18" / Int32ul,
            "addr3" / Int64ul,
            "addr4" / Int64ul,
            "unk_2c" / Int32ul,
            "unk_30" / Int32ul,
            "unk_34" / Int32ul,
        )
        args = s_cmd_pmp_ctrl_set.build(dict(
                opcode=CISP_CMD_PMP_CTRL_SET,
                addr0=0x23b738010,
            addr1=0x23bc3c000,
            unk_18=0x401,
            addr3=0x23b73800c,
            addr4=0x23bc3c000,
            unk_2c=0x400,
            unk_30=0x1000,
            unk_34=0x1000,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x38,
            outsize=0x0,
            args=args,
        )
        return self.send(cmd)

    def cmd_start(self):
        s_cmd_start = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "unk_8" / Int32ul,
            "unk_c" / Int32ul,
        )
        args = s_cmd_start.build(dict(
                opcode=CISP_CMD_START,
                unk_8=0x0,
            unk_c=0x0,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0xc,
            outsize=0xc,
            args=args,
        )
        return self.send(cmd)

    def cmd_config_get(self):
        s_cmd_config_get = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "unk_8" / Int32ul,
            "unk_c" / Int32ul,
            "unk_10" / Int32ul,
            "unk_14" / Int32ul,
            "unk_18" / Int32ul,
        )
        args = s_cmd_config_get.build(dict(
                opcode=CISP_CMD_CONFIG_GET,
                unk_8=0x0,
            unk_c=0x0,
            unk_10=0x0,
            unk_14=0x0,
            unk_18=0x0,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x1c,
            outsize=0x1c,
            args=args,
        )
        def cb(): chexdump32(self.isp.ioread(self.cmd_iova, 0x20))
        """
        00000000  00000000 00000003 016e3600 00000001 0000000a 00000000 00000001 00000000
        """
        return self.send(cmd, cb=cb)

    def cmd_ch_info_get(self):
        s_cmd_ch_info_get = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
            "unk_10" / Int32ul,
            "unk_14" / Int32ul,
            "unk_18" / Int32ul,
        )
        args = s_cmd_ch_info_get.build(dict(
                opcode=CISP_CMD_CH_INFO_GET,
                chan=0x0,
            unk_c=0x0,
            unk_10=0x0,
            unk_14=0x0,
            unk_18=0x0,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x118,
            outsize=0x118,
            args=args,
        )
        def cb(): chexdump32(self.isp.ioread(self.cmd_iova, 0x120))
        """
        00000000  00000000 0000010d 00000000 07da0001 000300ac 00040007 00000005 00000001
        00000020  00000248 00000007 00000001 00000001 00000000 00000000 00000000 00000000
        00000040  00000000 00000000 00000000 00010000 00000001 00000000 00000004 00000010
        00000060  00000001 00000000 000044c0 00000040 00000001 00000002 00004000 00000040
        00000080  00000001 00000000 00000000 00000036 00000000 00000000 000f4240 43430000
        000000a0  32353032 38513033 59334a56 00303333 00000000 00000008 00000000 00000000
        000000c0  00000004 00000000 00000000 00000000 00000000 00000000 00000000 00ff0000
        000000e0  00000c00 00000000 0000001c 00000640 00000004 00000004 00000000 00000000
        00000100  00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000
        """
        return self.send(cmd, cb=cb)

    def cmd_ch_camera_config_get(self):
        s_cmd_ch_camera_config_get = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
            "unk_10" / Int32ul,
            "unk_14" / Int32ul,
            "unk_18" / Int32ul,
        )
        args = s_cmd_ch_camera_config_get.build(dict(
                opcode=CISP_CMD_CH_CAMERA_CONFIG_GET,
                chan=0x0,
            unk_c=0x0,
            unk_10=0x0,
            unk_14=0x0,
            unk_18=0x0,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0xdc,
            outsize=0xdc,
            args=args,
        )
        def cb(): chexdump32(self.isp.ioread(self.cmd_iova, 0xe0))
        """
        00000000  00000000 00000106 00000000 00000000 02e00510 02e00510 00000000 00001df8
        00000020  00000100 00000001 00000040 00000040 00000040 00000040 00000040 00000040
        00000040  00000003 00000040 00000040 00000005 00000000 00000528 00000001 00000000
        00000060  0249f000 00000006 00000007 00000000 00000009 000f4240 00000025 00000000
        00000080  00004000 00000014 00000015 00000000 00000000 00000000 00000510 000002e0
        000000a0  00010000 00000000 00000080 039f2990 0000001f 00000000 00000000 00000000
        000000c0  00000510 000002e0 00000100 00000000 00000000 00000510 000002e0 00000000
        """
        return self.send(cmd, cb=cb)

    def cmd_ch_camera_config_select(self):
        s_cmd_ch_camera_config_select = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
            "unk_10" / Int32ul,
            "unk_14" / Int32ul,
            "unk_18" / Int32ul,
        )
        args = s_cmd_ch_camera_config_select.build(dict(
                opcode=CISP_CMD_CH_CAMERA_CONFIG_SELECT,
                chan=0x0,
            unk_c=0x0,
            unk_10=0x0,
            unk_14=0x0,
            unk_18=0x0,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x10,
            outsize=0x10,
            args=args,
        )
        return self.send(cmd)

    def cmd_ch_sbs_enable(self):
        s_cmd_ch_sbs_enable = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
        )
        args = s_cmd_ch_sbs_enable.build(dict(
                opcode=CISP_CMD_CH_SBS_ENABLE,
                chan=0x0,
            unk_c=0x1,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x10,
            outsize=0x10,
            args=args,
        )
        return self.send(cmd)

    def cmd_ch_buffer_recycle_mode_set(self):
        # ISPCPU: b'[MSC] CH = 0x0   Dynamic Buffers Recycling Mode Set [EMPTY ONLY] \n']
        s_cmd_ch_buffer_recycle_mode_set = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
        )
        args = s_cmd_ch_buffer_recycle_mode_set.build(dict(
                opcode=CISP_CMD_CH_BUFFER_RECYCLE_MODE_SET,
                chan=0x0,
            unk_c=0x1,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x10,
            outsize=0x10,
            args=args,
        )
        return self.send(cmd)

    def cmd_ch_buffer_recycle_start(self):
        s_cmd_ch_buffer_recycle_start = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
        )
        args = s_cmd_ch_buffer_recycle_start.build(dict(
                opcode=CISP_CMD_CH_BUFFER_RECYCLE_START,
                chan=0x0,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0xc,
            outsize=0xc,
            args=args,
        )
        return self.send(cmd)

    def cmd_ch_buffer_pool_config_set(self):
        """
        00000000  00000000 00000117 00000000 00100008 00004640 00004640 00000000 00000000
        00000080  00000000 00000000 00000000 00000000 00000000 00000001 00000000 000002e0
        TRC: CDSControllerBase.cpp, BufferPoolConfig, 2937:  0, type=8, count=16, id=8, compress=0, dataBlocks=1
        """
        s_cmd_ch_buffer_pool_config_set = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
            "unk_10" / Int32ul,
            "unk_14" / Int32ul,
            "zero" / Padding(0x7c),
            "dataBlocks" / Int32ul,
        )
        args = s_cmd_ch_buffer_pool_config_set.build(dict(
                opcode=CISP_CMD_CH_BUFFER_POOL_CONFIG_SET,
                chan=0x0,
            unk_c=0x100008,
            unk_10=0x4640,
            unk_14=0x4640,
            dataBlocks=0x1,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x9c,
            outsize=0x9c,
            args=args,
        )
        return self.send(cmd)

    def cmd_ch_buffer_recycle_stop(self):
        s_cmd_ch_buffer_recycle_stop = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
        )
        args = s_cmd_ch_buffer_recycle_stop.build(dict(
                opcode=CISP_CMD_CH_BUFFER_RECYCLE_STOP,
                chan=0x0,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0xc,
            outsize=0xc,
            args=args,
        )
        return self.send(cmd)

    def cmd_ch_camera_agile_freq_array_current_get(self):
        s_cmd_ch_camera_agile_freq_array_current_get = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
        )
        args = s_cmd_ch_camera_agile_freq_array_current_get.build(dict(
                opcode=CISP_CMD_CH_CAMERA_AGILE_FREQ_ARRAY_CURRENT_GET,
                chan=0x0,
            unk_c=0x0,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x10,
            outsize=0x10,
            args=args,
        )
        def cb(): chexdump32(self.isp.ioread(self.cmd_iova, 0x20))
        return self.send(cmd, cb=cb)

    def cmd_ch_crop_set(self):
        """
        00000000  00000000 00000801 00000000 00000008 00000008 00000500 000002d0 00000001
        00000020  00000248 00000007 00000001 00000001 00000000 00000000 00000000 00000000
        ISPCPU: b'[MSC] CH = 0 BES[0] Crop [8 8 1280 720] FES output [1296 736]\n']
        ISPCPU: b'[MSC] CH = 0x0  BES[0] CROP -> [8, 8][1280, 720] within [0, 0][1296, 736]  original crop = [8 8 1280 720]\n']
        """
        s_cmd_ch_crop_set = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
            "unk_10" / Int32ul,
            "unk_14" / Int32ul,
            "unk_18" / Int32ul,
            "unk_1c" / Int32ul,
        )
        args = s_cmd_ch_crop_set.build(dict(
                opcode=CISP_CMD_CH_CROP_SET,
                chan=0x0,
            unk_c=0x8,     # 8
            unk_10=0x8,    # 8
            unk_14=0x500,  # 1280
            unk_18=0x2d0,  # 720
            unk_1c=0x1,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x1c,
            outsize=0x1c,
            args=args,
        )
        return self.send(cmd)

    def cmd_ch_output_config_set(self):
        """
        00000000  00000000 00000b01 00000000 00000500 000002d0 00000001 00000000 00000500
        00000020  00000500 00000000 00000000 000002d0 00000000 00000500 00000000 00000000
        [MSC] CH = 0x0 Scl=0 Output Config: format=0,range=1,size=1280x720,paddingRows=0,cmpEn=0
        """
        s_cmd_ch_crop_set = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
            "unk_10" / Int32ul,
            "unk_14" / Int32ul,
            "unk_18" / Int32ul,
            "unk_1c" / Int32ul,
            "unk_20" / Int32ul,
            "unk_24" / Int32ul,
            "unk_28" / Int32ul,
            "unk_2c" / Int32ul,
            "unk_30" / Int32ul,
            "unk_34" / Int32ul,
            "unk_38" / Int32ul,
        )
        args = s_cmd_ch_crop_set.build(dict(
                opcode=CISP_CMD_CH_OUTPUT_CONFIG_SET,
                chan=0x0,

            unk_c=0x500,
            unk_10=0x2d0,
            unk_14=0x1,
            unk_18=0x0,

            unk_1c=0x500,
            unk_20=0x500,
            unk_24=0x0,
            unk_28=0x0,

            unk_2c=0x2d0,
            unk_30=0x0,
            unk_34=0x500,
            unk_38=0x0,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x38,
            outsize=0x38,
            args=args,
        )
        return self.send(cmd)

    def cmd_ch_preview_stream_set(self):
        # [MSC] CH = 0x0   Preview stream set = 1
        s_cmd_ch_preview_stream_set = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
        )
        args = s_cmd_ch_preview_stream_set.build(dict(
                opcode=CISP_CMD_CH_PREVIEW_STREAM_SET,
                chan=0x0,
            unk_c=0x1,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x10,
            outsize=0x10,
            args=args,
        )
        return self.send(cmd)

    def cmd_ch_cnr_start(self):
        s_cmd_ch_cnr_start = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
        )
        args = s_cmd_ch_cnr_start.build(dict(
                opcode=CISP_CMD_CH_CNR_START,
                chan=0x0,
            unk_c=0x1,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0xc,
            outsize=0xc,
            args=args,
        )
        return self.send(cmd)

    def cmd_ch_mbnr_enable(self):
        """
        00000000  00000000 00000a3a 00000000 00000000 00000001 00000001
        ISPCPU: [MSC] CH = 0, mbnrMode = 1,useCase = 0, enableChroma = 1
        """
        s_cmd_ch_mbnr_enable = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
            "unk_10" / Int32ul,
            "unk_14" / Int32ul,
        )
        args = s_cmd_ch_mbnr_enable.build(dict(
                opcode=CISP_CMD_CH_MBNR_ENABLE,
                chan=0x0,
            unk_c=0x0,
            unk_10=0x1,
            unk_14=0x1,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x18,
            outsize=0x18,
            args=args,
        )
        return self.send(cmd)

    def cmd_apple_ch_temporal_filter_start(self):
        """
        00000000  00000000 0000c100 00000000 00000001 00000000 00000001 00000000 00000500
        [MSC] CH = 0x0   bMSTFScale0En=1, fusionType=0, isStreaming=0
        """
        s_cmd_apple_ch_temporal_filter_start = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
            "unk_10" / Int32ul,
            "unk_14" / Int32ul,
        )
        args = s_cmd_apple_ch_temporal_filter_start.build(dict(
                opcode=CISP_CMD_APPLE_CH_TEMPORAL_FILTER_START,
                chan=0x0,
            unk_c=0x1,
            unk_10=0x0,
            unk_14=0x1,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x14,
            outsize=0x14,
            args=args,
        )
        return self.send(cmd)

    def cmd_apple_ch_motion_history_start(self):
        s_cmd_apple_ch_motion_history_start = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
        )
        args = s_cmd_apple_ch_motion_history_start.build(dict(
                opcode=CISP_CMD_APPLE_CH_MOTION_HISTORY_START,
                chan=0x0,
            unk_c=0x1,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0xc,
            outsize=0xc,
            args=args,
        )
        return self.send(cmd)

    def cmd_apple_ch_temporal_filter_enable(self):
        s_cmd_apple_ch_temporal_filter_enable = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
        )
        args = s_cmd_apple_ch_temporal_filter_enable.build(dict(
                opcode=CISP_CMD_APPLE_CH_TEMPORAL_FILTER_ENABLE,
                chan=0x0,
            unk_c=0x1,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0xc,
            outsize=0xc,
            args=args,
        )
        return self.send(cmd)

    def cmd_apple_ch_ae_fd_scene_metering_config_set(self):
        """
        00000000  00000000 0000820e 00000000 000000b8 02000200 00280800 00e10028 000a0399
        00000020  03cc02cc 00000000 00000000 000002d0 00000000 00000500 00000000 00000000
        ISPCPU: scene: T=184 low/hiK:0/512 sc:512..2048 outl:40,40 Q:10,716,972
        ISPCPU: face: T:225 maxFW:921
        """
        s_cmd_apple_ch_ae_fd_scene_metering_config_set = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
            "unk_10" / Int32ul,
            "unk_14" / Int32ul,
            "unk_18" / Int32ul,
            "unk_1c" / Int32ul,
            "unk_20" / Int32ul,
            "unk_24" / Int32ul,
            "unk_28" / Int32ul,
        )
        args = s_cmd_apple_ch_ae_fd_scene_metering_config_set.build(dict(
                opcode=CISP_CMD_APPLE_CH_AE_FD_SCENE_METERING_CONFIG_SET,
                chan=0x0,
            unk_c=0xb8,
            unk_10=0x2000200,
            unk_14=0x280800,
            unk_18=0xe10028,
            unk_1c=0xa0399,
            unk_20=0x3cc02cc,
            unk_24=0x0,
            unk_28=0x0,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x24,
            outsize=0x24,
            args=args,
        )
        return self.send(cmd)

    def cmd_apple_ch_ae_metering_mode_set(self):
        s_cmd_apple_ch_ae_metering_mode_set = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "mode" / Int32ul,
        )
        args = s_cmd_apple_ch_ae_metering_mode_set.build(dict(
                opcode=CISP_CMD_APPLE_CH_AE_METERING_MODE_SET,
                chan=0x0,
            mode=0x3,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x10,
            outsize=0x10,
            args=args,
        )
        return self.send(cmd)

    def cmd_ch_ae_stability_set(self):
        s_cmd_ch_ae_stability_set = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "stability" / Int32ul,
        )
        args = s_cmd_ch_ae_stability_set.build(dict(
                opcode=CISP_CMD_CH_AE_STABILITY_SET,
                chan=0x0,
            stability=0x20,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x10,
            outsize=0x10,
            args=args,
        )
        return self.send(cmd)

    def cmd_ch_ae_stability_to_stable_set(self):
        s_cmd_ch_ae_stability_to_stable_set = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "stability" / Int32ul,
        )
        args = s_cmd_ch_ae_stability_to_stable_set.build(dict(
                opcode=CISP_CMD_CH_AE_STABILITY_TO_STABLE_SET,
                chan=0x0,
            stability=0x14,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x10,
            outsize=0x10,
            args=args,
        )
        return self.send(cmd)

    def cmd_ch_sif_pixel_format_set(self):
        """
        00000000  00000000 00000115 00000000 00000103 00000000
        [MSC] CH = 0x0   Sif Pixel Format3, type 1 DmaCompress 0 Companding 0
        """
        s_cmd_ch_sif_pixel_format_set = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
            "unk_10" / Int32ul,
        )
        args = s_cmd_ch_sif_pixel_format_set.build(dict(
                opcode=CISP_CMD_CH_SIF_PIXEL_FORMAT_SET,
                chan=0x0,
            unk_c=0x103,
            unk_10=0x0,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x14,
            outsize=0x14,
            args=args,
        )
        return self.send(cmd)

    def cmd_ch_face_detection_config_get(self):
        s_cmd_ch_face_detection_config_get = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
            "unk_10" / Int32ul,
            "unk_14" / Int32ul,
            "unk_18" / Int32ul,
        )
        args = s_cmd_ch_face_detection_config_get.build(dict(
                opcode=CISP_CMD_CH_FACE_DETECTION_CONFIG_GET,
                chan=0x0,
            unk_c=0x103,
            unk_10=0x0,
            unk_14=0x0,
            unk_18=0x0,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x1c,
            outsize=0x1c,
            args=args,
        )
        def cb(): chexdump32(self.isp.ioread(self.cmd_iova, 0x20))
        """
        00000000  00000000 00000d02 00000000 00000000 0000000a 0000000a 00000000 00000000
        """
        return self.send(cmd, cb=cb)

    def cmd_ch_face_detection_config_set(self):
        """
        00000000  00000000 00000d03 00000000 0000000a 01000000 00000001
        MSG: : FDConfig Eye 0 Blink 0 Smile 0 nFace 10
        CISP_CMD_CH_FACE_DETECTION_CONFIG_SET enableAttr = 0
        enableOd = 0
        enableSaliency = 0, enableSaliencyHW = 1
        """
        s_cmd_ch_face_detection_config_set = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
            "unk_10" / Int32ul,
            "unk_14" / Int32ul,
        )
        args = s_cmd_ch_face_detection_config_set.build(dict(
                opcode=CISP_CMD_CH_FACE_DETECTION_CONFIG_SET,
                chan=0x0,
            unk_c=0xa,
            unk_10=0x1000000,
            unk_14=0x1,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x18,
            outsize=0x18,
            args=args,
        )
        return self.send(cmd)

    def cmd_ch_face_detection_enable(self):
        s_cmd_ch_face_detection_enable = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "FDEnableMask" / Int32ul,
        )
        args = s_cmd_ch_face_detection_enable.build(dict(
                opcode=CISP_CMD_CH_FACE_DETECTION_ENABLE,
                chan=0x0,
            FDEnableMask=0x1,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x10,
            outsize=0x10,
            args=args,
        )
        return self.send(cmd)

    def cmd_ch_face_detection_start(self):
        s_cmd_ch_face_detection_start = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
        )
        args = s_cmd_ch_face_detection_start.build(dict(
                opcode=CISP_CMD_CH_FACE_DETECTION_START,
                chan=0x0,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0xc,
            outsize=0xc,
            args=args,
        )
        return self.send(cmd)

    def cmd_ch_camera_config_current_get(self):
        s_cmd_ch_camera_config_current_get = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
            "unk_10" / Int32ul,
            "unk_14" / Int32ul,
            "unk_18" / Int32ul,
        )
        args = s_cmd_ch_camera_config_current_get.build(dict(
                opcode=CISP_CMD_CH_CAMERA_CONFIG_CURRENT_GET,
                chan=0x0,
            unk_c=0x0,
            unk_10=0x0,
            unk_14=0x0,
            unk_18=0x0,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0xdc,
            outsize=0xdc,
            args=args,
        )
        def cb(): chexdump32(self.isp.ioread(self.cmd_iova, 0xe0))
        """
        00000000  00000000 00000105 00000000 00000000 02e00510 02e00510 00000000 00001df8
        00000020  00000100 00000001 00000040 00000040 00000040 00000040 00000040 00000040
        00000040  00000003 00000040 00000040 00000005 00000000 00000528 00000001 00000000
        00000060  0249f000 00000006 00000007 00000000 00000009 000f4240 00000025 00000000
        00000080  00004000 00000014 00000015 00000000 00000000 00000000 00000510 000002e0
        000000a0  00010000 00000000 00000000 00000000 0000001f 00000000 00000000 00000000
        000000c0  00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000
        """
        return self.send(cmd, cb=cb)

    def cmd_ch_ae_frame_rate_max_set(self):
        # TODO IMPORTANT DONT FORGET EILEEN !!!!!!!!!!!
        # Normal framerate set by macos is 0x1e00 for both min/max.
        # Since m1n1 is too slow to keep up, temporarily using 1/8th, 0x3c0
        s_cmd_ch_ae_frame_rate_max_set = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
        )
        args = s_cmd_ch_ae_frame_rate_max_set.build(dict(
                opcode=CISP_CMD_CH_AE_FRAME_RATE_MAX_SET,
                chan=0x0,
            unk_c=0x3c0,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x10,
            outsize=0x10,
            args=args,
        )
        return self.send(cmd)

    def cmd_ch_ae_frame_rate_min_set(self):
        s_cmd_ch_ae_frame_rate_min_set = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
            "unk_c" / Int32ul,
        )
        args = s_cmd_ch_ae_frame_rate_min_set.build(dict(
                opcode=CISP_CMD_CH_AE_FRAME_RATE_MIN_SET,
                chan=0x0,
            unk_c=0x3c0,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0x10,
            outsize=0x10,
            args=args,
        )
        return self.send(cmd)

    def cmd_ch_start(self):  # green light :)
        s_cmd_ch_start = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
        )
        args = s_cmd_ch_start.build(dict(
                opcode=CISP_CMD_CH_START,
                chan=0x0,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0xc,
            outsize=0xc,
            args=args,
        )
        return self.send(cmd)

    def cmd_ch_stop(self):  # no more green light
        s_cmd_ch_stop = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "chan" / Int32ul,
        )
        args = s_cmd_ch_stop.build(dict(
                opcode=CISP_CMD_CH_STOP,
                chan=0x0,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0xc,
            outsize=0xc,
            args=args,
        )
        return self.send(cmd)

    def cmd_stop(self):
        s_cmd_stop = Struct(
                "pad" / Default(Int32ul, 0),
                "opcode" / Int32ul,
                "unk_c" / Int32ul,
        )
        args = s_cmd_stop.build(dict(
                opcode=CISP_CMD_STOP,
                unk_c=0x0,
        ))
        cmd = ISPIORequestCommand(
            iova=self.cmd_iova,
            insize=0xc,
            outsize=0xc,
            args=args,
        )
        return self.send(cmd)



from m1n1.trace import ADTDevTracer
from m1n1.trace.dart8110 import DART8110Tracer
from m1n1.hw.prores import *
from m1n1.utils import *
import struct

p.pmgr_adt_clocks_enable('/arm-io/dart-apr0')
p.pmgr_adt_clocks_enable('/arm-io/dart-apr1')

dart0_tracer = DART8110Tracer(hv, "/arm-io/dart-apr0", verbose=1)
dart0_tracer.start()
print(dart0_tracer)
dart1_tracer = DART8110Tracer(hv, "/arm-io/dart-apr1", verbose=1)
dart1_tracer.start()
print(dart1_tracer)


class ProResTracer(ADTDevTracer):
    DEFAULT_MODE = TraceMode.SYNC
    REGMAPS = [ProResRegs]
    NAMES = ['prores']

    def __init__(self, hv, devpath, dart_tracer):
        super().__init__(hv, devpath, verbose=3)
        self.dart_tracer = dart_tracer

    def w_DR_SIZE(self, val):
        self._dr_size = val

    def w_DR_ADDR_HI(self, val):
        self._dr_addr_hi = val

    def w_DR_ADDR_LO(self, val):
        self._dr_addr_lo = val

    def w_DR_TAIL(self, val):
        self.log(f"DR_TAIL = {val}")
        self._dr_tail = val

    def w_DR_HEAD(self, val):
        self.log(f"DR_HEAD = {val}")
        self.dart_tracer.dart.dump_all()

        dr_addr = int(self._dr_addr_hi) << 32 | int(self._dr_addr_lo)
        dr_size = int(self._dr_size)
        self.log(f"desc ring @ {dr_addr:016X} sz {dr_size:08X}")

        dr = self.dart_tracer.dart.ioread(0, dr_addr, dr_size)
        chexdump(dr)

        # FIXME there are other descriptor types
        # also, what if there are multiple in the ring?
        dr_head = int(val)
        dr_tail = int(self._dr_tail)

        if dr_head - dr_tail == 0x180:
            desc = EncodeNotRawDescriptor._make(struct.unpack(ENCODE_NOT_RAW_STRUCT, dr[dr_tail:dr_head]))
            print(desc)

            p0_iova = desc.luma_iova
            p1_iova = desc.chroma_iova
            p2_iova = desc.alpha_iova

            if p0_iova:
                print(f"P0 iova {p0_iova:016X}")
                data = self.dart_tracer.dart.ioread(0, p0_iova, 0x1000)
                chexdump(data)
            if p1_iova:
                print(f"P1 iova {p1_iova:016X}")
                data = self.dart_tracer.dart.ioread(0, p1_iova, 0x1000)
                chexdump(data)
            if p2_iova:
                print(f"P2 iova {p2_iova:016X}")
                data = self.dart_tracer.dart.ioread(0, p2_iova, 0x1000)
                chexdump(data)


ProResTracer = ProResTracer._reloadcls()

p.pmgr_adt_clocks_enable('/arm-io/apr0')
p.pmgr_adt_clocks_enable('/arm-io/apr1')

tracer0 = ProResTracer(hv, '/arm-io/apr0', dart0_tracer)
tracer0.start()
print(tracer0)
tracer1 = ProResTracer(hv, '/arm-io/apr1', dart1_tracer)
tracer1.start()
print(tracer1)



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib, time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1 import asm

LOOPS = 10000000
freq = u.mrs(CNTFRQ_EL0)

CREG = [
    0x210e00000,
    0x211e00000,
]

CLUSTER_PSTATE = 0x20020

# e-core pstates
#   600 972 1332 1704 2064
# p-core pstates
#   600 828 1056 1284 1500 1728 1956 2184 2388 2592 2772 2988 3096 3144 3204

code = u.malloc(0x1000)

util = asm.ARMAsm("""
bench:
    mrs x1, CNTPCT_EL0
1:
    sub x0, x0, #1
    cbnz x0, 1b

    mrs x2, CNTPCT_EL0
    sub x0, x2, x1
    ret
""", code)
iface.writemem(code, util.data)
p.dc_cvau(code, len(util.data))
p.ic_ivau(code, len(util.data))

def bench_cpu(idx):
    if idx == 0:
        elapsed = p.call(util.bench, LOOPS) / freq
    else:
        elapsed = p.smp_call_sync(idx, util.bench, LOOPS) / freq
    if elapsed == 0:
        return 0
    mhz = (LOOPS / elapsed) / 1000000
    return mhz

print()

e_pstate = p.read64(CREG[0] + CLUSTER_PSTATE)
p_pstate = p.read64(CREG[1] + CLUSTER_PSTATE)

print(f"E-Core pstate: {e_pstate:x}")
print(f"P-Core pstate: {p_pstate:x}")

#for cluster in range(2):
    #print(f"Initializing cluster {cluster} (early)")
    
    #p.write64(CREG[cluster] + 0x20660, 0x1000000015)
    #p.write64(CREG[cluster] + 0x48000, 0)
    #p.write64(CREG[cluster] + 0x48080, 0xa000000000000000)

    #p.clear64(CREG[cluster] + CLUSTER_PSTATE, 1<<22)

#p.set32(PMGR + 0x48000, 1)
#p.set32(PMGR + 0x48c00, 1)
#p.set32(PMGR + 0x48800, 1)
#p.set32(PMGR + 0x48400, 1)

CLUSTER_DVMR = 0x206b8
CLUSTER_LIMIT2 = 0x40240
CLUSTER_LIMIT3 = 0x40250
CLUSTER_LIMIT1 = 0x48400

PMGR_CPUGATING = 0x1c080
CLUSTER_CTRL = 0x440f8
CLUSTER_PSCTRL = 0x200f8

for cluster in range(2):
    print(f"Initializing cluster {cluster}")
    ena = (1<<63)
    val = p.read64(CREG[cluster] + CLUSTER_DVMR)
    if cluster == 1:
        ena |= (1<<32) | (1<<31)
    if (val & ena) != ena:
        print(f"DVMR: {val:#x} -> {val|ena:#x}")
        p.set64(CREG[cluster] + CLUSTER_DVMR, ena) # CLUSTER_DVMR
    
    #p.set64(CREG[cluster] + CLUSTER_LIMIT1, 1<<63)
    #p.clear64(CREG[cluster] + CLUSTER_LIMIT2, 1<<63)
    #p.set64(CREG[cluster] + CLUSTER_LIMIT3, 1<<63)
    
    #p.set64(CREG[cluster] + CLUSTER_PSTATE, 0)
    
    #p.set32(PMGR + PMGR_CPUGATING + 8 * cluster, 1<<31)

    #p.write64(CREG[cluster] + CLUSTER_CTRL, 1)
    
    #p.set64(CREG[cluster] + CLUSTER_PSCTRL, 1<<40)

    #pstate = p.read64(CREG[cluster] + CLUSTER_PSTATE) & 0xf

p.smp_start_secondaries()

print("== Initial CPU frequencies ==")

for cpu in range(8):
    print(f"CPU {cpu}: {bench_cpu(cpu):.2f} MHz")
    
def set_pstate(cluster, pstate):
    # This really seems to be all that's needed
    
    p.mask64(CREG[cluster] + CLUSTER_PSTATE, 0xf00f, (1<<25) | pstate | (pstate << 12))

    # Optionally, adjust MCC performance in higher p-core pstates
    if cluster == 1:
        if pstate > 8:
            p0, p1 = 0x133, 0x55555340
        else:
            p0, p1 = 0x813057f, 0x1800180

        for lane in range(8):
            p.write32(0x200200dc4 + lane * 0x40000, p0)
            p.write32(0x200200dbc + lane * 0x40000, p1)

    # This seems to be about notifying PMP
    #p.write32(0x23b738004 + cluster*4, pstate)
    #p.write32(0x23bc34000, 1 << cluster)

set_pstate(1, 15)

e_pstate = p.read64(CREG[0] + CLUSTER_PSTATE)
p_pstate = p.read64(CREG[1] + CLUSTER_PSTATE)

print(f"E-Core pstate: {e_pstate:x}")
print(f"P-Core pstate: {p_pstate:x}")

time.sleep(0.5)

print("== Final CPU frequencies ==")

#elapsed = p.smp_call(7, util.bench, 80000000)

for cpu in range(8):
    print(f"CPU {cpu}: {bench_cpu(cpu):.2f} MHz")

#elapsed = p.smp_wait(7)




import random

from m1n1.utils import *
from m1n1.constructutils import *
from construct import *
from .cmdqueue import *

__all__ = ["channelNames", "channelRings", "DeviceControlMsg", "EventMsg", "StatsMsg", "StatsSize"]

if Ver.check("G >= G14 && V >= V13_2 && G < G14X"):
    RunCmdQueueSize = 0x40
else:
    RunCmdQueueSize = 0x30

class RunCmdQueueMsg(ConstructClass):
    subcon = Struct (
        "queue_type" / Default(Int32ul, 0),
        "cmdqueue_addr" / Default(Hex(Int64ul), 0),
        "cmdqueue" / Lazy(ROPointer(this.cmdqueue_addr, CommandQueueInfo)),
        "head" / Default(Int32ul, 0),
        "event_number" / Default(Int32ul, 0),
        "new_queue" / Default(Int32ul, 0),
        "timestamp" / Default(Int64ul, 0),
        "data" / HexDump(Default(Bytes(0x10), bytes(0x10))),
        Ver("G >= G14 && V >= V13_2 && G < G14X", ZPadding(0x10)),
    )

    TYPES = {
        0: "SubmitTA",
        1: "Submit3D",
        2: "SubmitCompute",
    }

    def __str__(self, *args, **kwargs):
        s = super().__str__(*args, **kwargs) + "\n"

        if self.cmdqueue_addr == 0:
            return s + "<Empty RunCmdQueueMsg>"

        r = random.randrange(2**64)
        s += f"{self.TYPES[self.queue_type]}(0x{self.cmdqueue_addr & 0xfff_ffffffff:x}, {self.head}, ev={self.event_number}, new={self.new_queue}) //{r:x}"
        return s

class DC_DestroyContext(ConstructClass):
    subcon =  Struct (
        Ver("V < V13_3", "msg_type" / Const(0x17, Int32ul)),
        Ver("V >= V13_3", "msg_type" / Const(0x18, Int32ul)),
        "unk_4" / Hex(Int32ul),
        "unk_8" / Hex(Int32ul),
        "unk_c" / Hex(Int32ul),
        "unk_10" / Hex(Int32ul),
        "unk_14" / Hex(Int32ul),
        "unk_18" / Hex(Int32ul),
        "context_addr" / Hex(Int64ul),
        "rest" / HexDump(Default(Bytes(0xc), bytes(0xc))),
        Ver("G == G14 && V >= V13_2", ZPadding(0x10)),
    )

class DC_Write32(ConstructClass):
    subcon =  Struct (
        Ver("V < V13_3", "msg_type" / Const(0x18, Int32ul)),
        Ver("V >= V13_3", "msg_type" / Const(0x19, Int32ul)),
        "addr" / Hex(Int64ul),
        "data" / Int32ul,
        "unk_10" / Int32ul,
        "unk_14" / Int32ul,
        "unk_18" / Int32ul,
        "unk_1c" / Int32ul,
        "rest" / HexDump(Default(Bytes(0x10), bytes(0x10))),
        Ver("G == G14 && V >= V13_2", ZPadding(0x10)),
    )

class DC_Write32B(ConstructClass):
    subcon =  Struct (
        "msg_type" / Const(0x13, Int32ul),
        "addr" / Hex(Int64ul),
        "data" / Int32ul,
        "unk_10" / Int32ul,
        "unk_14" / Int32ul,
        "unk_18" / Int32ul,
        "unk_1c" / Int32ul,
        "rest" / HexDump(Default(Bytes(0x10), bytes(0x10))),
        Ver("G == G14 && V >= V13_2", ZPadding(0x10)),
    )

class DC_Init(ConstructClass):
    subcon =  Struct (
        Ver("V < V13_3", "msg_type" / Const(0x19, Int32ul)),
        Ver("V >= V13_3", "msg_type" / Const(0x1a, Int32ul)),
        "data" / HexDump(Default(Bytes(0x2c), bytes(0x2c))),
        Ver("G == G14 && V >= V13_2", ZPadding(0x10)),
    )

class DC_09(ConstructClass):
    subcon =  Struct (
        "msg_type" / Const(0x9, Int32ul),
        "unk_4" / Int64ul,
        "unkptr_c" / Int64ul,
        "unk_14" / Int64ul,
        "data" /  HexDump(Default(Bytes(0x14), bytes(0x14))),
        Ver("G == G14 && V >= V13_2", ZPadding(0x10)),
    )

class DC_ConfigurePerfCounters(ConstructClass):
    subcon = Struct (
        "msg_type" / Const(0xb, Int32ul),
        "data" / HexDump(Default(Bytes(0x2c), bytes(0x2c))),
        Ver("G == G14 && V >= V13_2", ZPadding(0x10)),
    )

class DC_DisablePerfCounters(ConstructClass):
    subcon = Struct (
        "msg_type" / Const(0xc, Int32ul),
        "data" / HexDump(Default(Bytes(0x2c), bytes(0x2c))),
        Ver("G == G14 && V >= V13_2", ZPadding(0x10)),
    )

class DC_GrowTVBAck(ConstructClass):
    subcon =  Struct (
        "msg_type" / Const(0xd, Int32ul),
        "unk_4" / Int32ul,
        "bm_id" / Int32ul,
        "vm_id" / Int32ul,
        "counter" / Int32ul,
        "rest" / HexDump(Default(Bytes(0x1c), bytes(0x1c))),
        Ver("G == G14 && V >= V13_2", ZPadding(0x10)),
    )

class DC_Any(ConstructClass):
    subcon =  Struct (
        "msg_type" / Int32ul,
        "data" / HexDump(Default(Bytes(0x2c), bytes(0x2c))),
        Ver("G == G14 && V >= V13_2", ZPadding(0x10)),
    )

class DC_1e(ConstructClass):
    subcon =  Struct (
        "msg_type" / Const(0x1e, Int32ul),
        "unk_4" / Int64ul,
        "unk_c" / Int64ul,
        "data" /  HexDump(Default(Bytes(0x1c), bytes(0x1c))),
        Ver("G == G14 && V >= V13_2", ZPadding(0x10)),
    )

class DC_UpdateIdleTS(ConstructClass):
    subcon = Struct (
        "msg_type" / Const(0x23, Int32ul),
        "data" / HexDump(Default(Bytes(0x2c), bytes(0x2c))),
        Ver("G == G14 && V >= V13_2", ZPadding(0x10)),
    )

class UnknownMsg(ConstructClass):
    subcon = Struct (
        "msg_type" / Hex(Int32ul),
        "data" / HexDump(Bytes(0x2c)),
        Ver("G == G14 && V >= V13_2", ZPadding(0x10)),
    )

if Ver.check("G == G14 && V >= V13_2"):
    DeviceControlSize = 0x40
else:
    DeviceControlSize = 0x30

DeviceControlMsg = FixedSized(DeviceControlSize, Select(
    DC_DestroyContext,
    DC_Init,
    DC_UpdateIdleTS,
    DC_1e,
    DC_Write32,
    DC_ConfigurePerfCounters,
    DC_DisablePerfCounters,
    DC_GrowTVBAck,
    UnknownMsg,
))

class StatsMsg_Power(ConstructClass):
    subcon = Struct (
        "msg_type" / Hex(Const(0x00, Int32ul)),
        ZPadding(0x18), # ??? why the hole? never written...
        "power" / Hex(Int64ul),
        ZPadding(0xc), # Confirmed padding
        Ver("V >= V13_0B4", ZPadding(0x10)),
    )

    def __str__(self):
        return f"Power: {self.power / 8192.0:.3f} mW"

class StatsMsg_PowerOn(ConstructClass):
    subcon = Struct (
        "msg_type" / Hex(Const(0x02, Int32ul)),
        "power_off_ticks" / Dec(Int64ul),
        ZPadding(0x24), # Confirmed padding
        Ver("V >= V13_0B4", ZPadding(0x10)),
    )
    def __str__(self):
        t = self.power_off_ticks / 24000000
        return f"Power ON: spent {t:.04}s powered off ({self.power_off_ticks} ticks)"

class StatsMsg_PowerOff(ConstructClass):
    subcon = Struct (
        "msg_type" / Hex(Const(0x03, Int32ul)),
        "power_on_ticks" / Dec(Int64ul),
        ZPadding(0x24), # Confirmed padding
        Ver("V >= V13_0B4", ZPadding(0x10)),
    )
    def __str__(self):
        t = self.power_on_ticks / 24000000
        return f"Power OFF: spent {t:.04}s powered on ({self.power_on_ticks} ticks)"

class StatsMsg_Util(ConstructClass):
    subcon = Struct (
        "msg_type" / Hex(Const(0x04, Int32ul)),
        "timestamp" / Hex(Int64ul),
        "util1" / Dec(Int32ul),
        "util2" / Dec(Int32ul),
        "util3" / Dec(Int32ul),
        "util4" / Dec(Int32ul),
        ZPadding(0x14), # Confirmed padding
        Ver("V >= V13_0B4", ZPadding(0x10)),
    )
    def __str__(self):
        return f"Utilization: {self.util1:>3d}% {self.util2:>3d}% {self.util3:>3d}% {self.util4:>3d}%"

class StatsMsg_AvgPower(ConstructClass):
    subcon = Struct (
        "msg_type" / Hex(Const(0x09, Int32ul)),
        "active_cs" / Dec(Int64ul),
        "unk2" / Hex(Int32ul),
        "unk3" / Hex(Int32ul),
        "unk4" / Hex(Int32ul),
        "avg_power" / Dec(Int32ul),
        ZPadding(0x14), # Confirmed padding
        Ver("V >= V13_0B4", ZPadding(0x10)),
    )

    def __str__(self):
        return f"Activity: Active {self.active_cs * 10:6d} ms Avg Pwr {self.avg_power:4d} mW ({self.unk2:d} {self.unk3:d} {self.unk4:d})"

class StatsMsg_Temp(ConstructClass):
    subcon = Struct (
        "msg_type" / Hex(Const(0x0a, Int32ul)),
        ZPadding(8), # Not written
        "raw_value" / Hex(Int32ul),
        "scale" / Hex(Int32ul),
        "tmin" / Hex(Int32ul),
        "tmax" / Hex(Int32ul),
        ZPadding(0x14), # Confirmed padding
        Ver("V >= V13_0B4", ZPadding(0x10)),
    )

    def __str__(self):
        temp = self.raw_value / float(self.scale) / 64.0
        return f"Temp: {temp:.2f}°C s={self.scale:d} tmin={self.tmin:d} tmax={self.tmax:d}"

class StatsMsg_PowerState(ConstructClass):
    subcon = Struct (
        "msg_type" / Hex(Const(0x0b, Int32ul)),
        "timestamp" / Hex(Int64ul),
        "last_busy_ts" / Hex(Int64ul),
        "active" / Hex(Int32ul),
        "poweroff" / Dec(Int32ul),
        "unk2" / Dec(Int32ul),
        "pstate" / Dec(Int32ul),
        "unk4" / Dec(Int32ul),
        "unk5" / Dec(Int32ul),
        ZPadding(4), # Confirmed padding
        Ver("V >= V13_0B4", ZPadding(0x10)),
    )

    def __str__(self):
        act = "ACT" if self.active else "   "
        off = "OFF" if self.poweroff else "   "

        return f"PowerState: {act} {off} ps={int(self.pstate)} {self.unk4} {self.unk2} {self.unk5}"

class StatsMsg_FWBusy(ConstructClass):
    subcon = Struct (
        "msg_type" / Hex(Const(0x0c, Int32ul)),
        "timestamp" / Hex(Int64ul),
        "flag" / Int32ul,
        ZPadding(0x20), # Confirmed padding
        Ver("V >= V13_0B4", ZPadding(0x10)),
    )

    def __str__(self):
        return f"FW active: {bool(self.flag)}"

class StatsMsg_PState(ConstructClass):
    subcon = Struct (
        "msg_type" / Hex(Const(0x0d, Int32ul)),
        ZPadding(8), # Not written
        "ps_min" / Dec(Int32ul),
        "unk1" / Dec(Int32ul),
        "ps_max" / Dec(Int32ul),
        "unk3" / Dec(Int32ul),
        ZPadding(0x14), # Confirmed padding
        Ver("V >= V13_0B4", ZPadding(0x10)),
    )
    def __str__(self):
        return f"PState: {self.ps_min:d}..{self.ps_max:d} ({self.unk1:d}/{self.unk3:d})"

class StatsMsg_TempSensor(ConstructClass):
    subcon = Struct (
        "msg_type" / Hex(Const(0x0e, Int32ul)),
        ZPadding(4), # Not written
        "sensor_id" / Hex(Int32ul),
        "raw_value" / Hex(Int32ul),
        "scale" / Dec(Int32ul),
        "tmin" / Dec(Int32ul),
        "tmax" / Dec(Int32ul),
        ZPadding(0x14), # Confirmed padding
        Ver("V >= V13_0B4", ZPadding(0x10)),
    )
    def __str__(self):
        temp = self.raw_value / float(self.scale) / 64.0
        return f"TempSensor: #{self.sensor_id:d} {temp:.2f}°C s={self.scale:d} tmin={self.tmin:d} tmax={self.tmax:d}"

if Ver.check("V < V13_0B4"):
    StatsSize = 0x30
else:
    StatsSize = 0x40

StatsMsg = FixedSized(StatsSize, Select(
    StatsMsg_Power,
    StatsMsg_PowerOn,
    StatsMsg_PowerOff,
    StatsMsg_Util,
    StatsMsg_AvgPower,
    StatsMsg_Temp,
    StatsMsg_PowerState,
    StatsMsg_FWBusy,
    StatsMsg_PState,
    StatsMsg_TempSensor,
    UnknownMsg,
))

class FWLogMsg(ConstructClass):
    subcon = Struct (
        "msg_type" / Hex(Const(0x03, Int32ul)),
        "seq_no" / Hex(Int32ul),
        "timestamp" / Hex(Int64ul),
        "msg" / PaddedString(0xc8, "ascii")
    )

class FaultMsg(ConstructClass):
    subcon = Struct (
        "msg_type" / Hex(Const(0, Int32ul)),
        "unk_4" / HexDump(Bytes(0x34)),
    )

class FlagMsg(ConstructClass):
    subcon = Struct (
        "msg_type" / Hex(Const(1, Int32ul)),
        "firing" / Array(2, Hex(Int64ul)),
        "unk_14" / Hex(Int16ul),
        "tail" / Bytes(0x38 - 0x18),
    )

class TimeoutMsg(ConstructClass):
    subcon = Struct (
        "msg_type" / Hex(Const(4, Int32ul)),
        "counter" / Hex(Int64ul),
        "stamp_index" / Hex(Int32sl),
        "unkpad_16" / HexDump(Bytes(0x38 - 0x10)),
    )

class GrowTVBMsg(ConstructClass):
    subcon = Struct (
        "msg_type" / Hex(Const(7, Int32ul)),
        "vm_id" / Hex(Int32ul),
        "bm_id" / Hex(Int32ul),
        "counter" / Hex(Int32ul),
        "tail" / HexDump(Bytes(0x38 - 0x10)),
    )

class ChannelError(ConstructClass):
    subcon = Struct (
        "msg_type" / Hex(Const(8, Int32ul)),
        "tail" / HexDump(Bytes(0x38 - 0x4)),
    )

EventMsg = FixedSized(0x38, Select(
    FaultMsg,
    FlagMsg,
    TimeoutMsg,
    GrowTVBMsg,
    ChannelError,
    HexDump(Bytes(0x38)),
))

TRACE_MSGS = {
    (0x00, 0x00, 0): ("StartTA", "uuid", None, "unk", "cmdqueue"),
    (0x00, 0x01, 0): ("FinishTA", "uuid", None, "unk", "cmdqueue"),
    (0x00, 0x04, 0): ("Start3D", "uuid", "partial_render", "unk", "cmdqueue"),
    (0x00, 0x05, 0): ("Finish3D_unk", "uuid", "unk", "flag", "buf_related"),
    (0x00, 0x06, 0): ("Finish3D", "uuid", None, "unk", "cmdqueue"),
    (0x00, 0x07, 0): ("StartCP", "uuid", None, "unk", "cmdqueue"),
    (0x00, 0x08, 0): ("FinishCP", "uuid", None, "unk", "cmdqueue"),
    (0x00, 0x0a, 0): ("StampUpdateTA", "value", "ev_id", "addr", "uuid"),
    (0x00, 0x0c, 0): ("StampUpdate3D", "value", "ev_id", "addr", "uuid"),
    (0x00, 0x0e, 0): ("StampUpdateCL", "value", "ev_id", "addr", "uuid"),
    (0x00, 0x10, 1): ("TAPreproc1", "unk"),
    (0x00, 0x10, 2): ("TAPreproc2", "unk1", "unk2"),
    (0x00, 0x17, 0): ("Finish3D2", "uuid", None, "unk", "cmdqueue"),
    (0x00, 0x28, 0): ("EvtNotify", "firing0", "firing1", "firing2", "firing3"),
    (0x00, 0x2f, 0): ("Finish3D_unk2", "uuid", "unk"),
    (0x00, 0x1e, 0): ("CleanupPB", "uuid", "unk2", "slot"),
    (0x01, 0x0a, 0): ("Postproc", "cmdid", "event_ctl", "stamp_val", "uuid"),
    (0x01, 0x0b, 0): ("EvtComplete", None, "event_ctl"),
    (0x01, 0x0d, 0): ("EvtDequeued", "next", "event_ctl"),
    (0x01, 0x16, 0): ("InitAttachment", "idx", "flags", "addr", "size"),
    (0x01, 0x18, 0): ("ReInitAttachment", "idx", "flags", "addr", "size"),
}

class KTraceMsg(ConstructClass):
    THREADS = [
        "irq",
        "bg",
        "smpl",
        "pwr",
        "rec",
        "kern",
    ]
    subcon = Struct (
        "msg_type" / Hex(Const(5, Int32ul)),
        "timestamp" / Hex(Int64ul),
        "args" / Array(4, Int64ul),
        "code" / Int8ul,
        "channel" / Int8ul,
        "pad" / Const(0, Int8ul),
        "thread" / Int8ul,
        "unk_flag" / Int64ul,
    )
    def __str__(self):
        ts = self.timestamp / 24000000
        code = (self.channel, self.code, self.unk_flag)
        if code in TRACE_MSGS:
            info = TRACE_MSGS[code]
            args = info[0] + ": " + " ".join(f"{k}={v:#x}" for k, v in zip(info[1:], self.args) if k is not None)
        else:
            args = "UNK: " + ", ".join(hex(i) for i in self.args)
        return f"TRACE: [{ts:10.06f}][{self.THREADS[self.thread]:4s}] {self.channel:2x}:{self.code:2x} ({self.unk_flag}) {args}"

class FWCtlMsg(ConstructClass):
    subcon = Struct (
        "addr" / Int64ul,
        "unk_8" / Int32ul,
        "context_id" / Int32ul,
        "unk_10" / Int16ul,
        "unk_12" / Int16ul,
    )

channelNames = [
    "TA_0", "3D_0", "CL_0",
    "TA_1", "3D_1", "CL_1",
    "TA_2", "3D_2", "CL_2",
    "TA_3", "3D_3", "CL_3",
    "DevCtrl",
    "Event", "FWLog", "KTrace", "Stats",

    ## Not really in normal order
    "FWCtl"
]

# Exclude FWCtl
CHANNEL_COUNT = len(channelNames) - 1

channelRings = (
    [[(RunCmdQueueMsg, RunCmdQueueSize, 0x100)]] * 12 + [
        [(DeviceControlMsg, DeviceControlSize, 0x100)],
        [(EventMsg, 0x38, 0x100)],
        [
            (FWLogMsg, 0xd8, 0x100),                # unk 0
            (FWLogMsg, 0xd8, 0x100),                # init log
            (FWLogMsg, 0xd8, 0x100),                # unk 2
            (FWLogMsg, 0xd8, 0x100),                # warnings?
            (FWLogMsg, 0xd8, 0x100),                # unk 4
            (FWLogMsg, 0xd8, 0x100),                # unk 5
        ],
        [(KTraceMsg, 0x38, 0x200)],
        [(StatsMsg, StatsSize, 0x100)],
        [(FWCtlMsg, 0x14, 0x100)],
    ]
)

class ChannelStateFields(RegMap):
    _SIZE = 0x30

    READ_PTR = 0x00, Register32
    WRITE_PTR = 0x20, Register32

class FWControlStateFields(RegMap):
    _SIZE = 0x20

    READ_PTR = 0x00, Register32
    WRITE_PTR = 0x10, Register32

class Channel(Reloadable):
    def __init__(self, u, uat, info, ring_defs, base=None, state_fields=ChannelStateFields):
        self.uat = uat
        self.u = u
        self.p = u.proxy
        self.iface = u.iface

        self.ring_defs = ring_defs
        self.info = info

        self.accessor = uat.ioaccessor(0)
        self.state_addr = info.state_addr
        self.state = []
        self.rb_base = []
        self.rb_maps = []

        if base is None:
            p = info.ringbuffer_addr
        else:
            p = base
        for i, (msg, size, count) in enumerate(ring_defs):
            assert msg.sizeof() == size

            self.state.append(state_fields(self.accessor, self.state_addr + 0x30 * i))
            m = uat.iotranslate(0, p, size * count)
            self.rb_base.append(p)
            self.rb_maps.append(m)
            p += size * count

    def get_message(self, ring, index, meta_fn=None):
        msgcls, size, count = self.ring_defs[ring]

        assert index < count
        addr = self.rb_base[ring] + index * size
        stream = self.uat.iostream(0, addr)
        if meta_fn is not None:
            stream.meta_fn = lambda a, b: meta_fn(0, a, b)
        return msgcls.parse_stream(stream)

    def clear_message(self, ring, index):
        msgcls, size, count = self.ring_defs[ring]

        self.put_message(ring, index, b"\xef\xbe\xad\xde" * (size // 4))

    def put_message(self, ring, index, obj):
        msgcls, size, count = self.ring_defs[ring]

        assert index < count
        if isinstance(obj, bytes):
            data = obj
        else:
            data = obj.build()
        self.uat.iowrite(0, self.rb_base[ring] + index * size, data)

class ChannelInfo(ConstructClass):
    subcon = Struct(
        "state_addr" / Hex(Int64ul),
        "ringbuffer_addr" / Hex(Int64ul),
    )

class ChannelInfoSet(ConstructClass):
    CHAN_COUNT = CHANNEL_COUNT

    subcon = Struct(*[ name / ChannelInfo for name in channelNames[:CHAN_COUNT]])

__all__.extend(k for k, v in globals().items()
               if (callable(v) or isinstance(v, type)) and v.__module__ == __name__)



# SPDX-License-Identifier: MIT
from contextlib import contextmanager

__all__ = ["Heap"]

class Heap(object):
    def __init__(self, start, end, block=64):
        if start%block:
            raise ValueError("heap start not aligned")
        if end%block:
            raise ValueError("heap end not aligned")
        self.offset = start
        self.count = (end - start) // block
        self.blocks = [(self.count,False)]
        self.block = block

    def malloc(self, size):
        size = (size + self.block - 1) // self.block
        pos = 0
        for i, (bsize, full) in enumerate(self.blocks):
            if not full and bsize >= size:
                self.blocks[i] = (size, True)
                if bsize > size:
                    self.blocks.insert(i+1, (bsize - size, False))
                return self.offset + self.block * pos
            pos += bsize
        raise Exception("Out of memory")

    def memalign(self, align, size):
        assert (align & (align - 1)) == 0
        align = max(align, self.block) // self.block
        size = (size + self.block - 1) // self.block
        pos = self.offset // self.block
        for i, (bsize, full) in enumerate(self.blocks):
            if not full:
                offset = 0
                if pos % align:
                    offset = align - (pos % align)
                if bsize >= (size + offset):
                    if offset:
                        self.blocks.insert(i, (offset, False))
                        i += 1
                    self.blocks[i] = (size, True)
                    if bsize > (size + offset):
                        self.blocks.insert(i+1, (bsize - size - offset, False))
                    return self.block * (pos + offset)
            pos += bsize
        raise Exception("Out of memory")

    def free(self, addr):
        if addr%self.block:
            raise ValueError("free address not aligned")
        if addr<self.offset:
            raise ValueError("free address before heap")
        addr -= self.offset
        addr //= self.block
        if addr>=self.count:
            raise ValueError("free address after heap")
        pos = 0
        for i, (bsize, used) in enumerate(self.blocks):
            if pos > addr:
                raise ValueError("bad free address")
            if pos == addr:
                if used == False:
                    raise ValueError("block already free")
                if i!=0 and self.blocks[i-1][1] == False:
                    bsize += self.blocks[i-1][0]
                    del self.blocks[i]
                    i -= 1
                if i!=(len(self.blocks)-1) and self.blocks[i+1][1] == False:
                    bsize += self.blocks[i+1][0]
                    del self.blocks[i]
                self.blocks[i] = (bsize, False)
                return
            pos += bsize
        raise ValueError("bad free address")

    def check(self):
        free = 0
        inuse = 0
        for i, (bsize, used) in enumerate(self.blocks):
            if used:
                inuse += bsize
            else:
                free += bsize
        if free + inuse != self.count:
            raise Exception("Total block size is inconsistent")
        print("Heap stats:")
        print(" In use: %8dkB"%(inuse * self.block // 1024))
        print(" Free:   %8dkB"%(free * self.block // 1024))

    @contextmanager
    def guarded_malloc(self, size):
        addr = self.malloc(size)
        try:
            yield addr
        finally:
            self.free(addr)



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *

p = 0x800000000
limit = u.base
block = 0x40000

while p < limit:
    f = "mem/0x%x.bin" % p
    if os.path.exists(f):
        p += block
        continue

    print("dumping 0x%x..." % p)

    data = iface.readmem(p, block)
    open(f, "wb").write(data)
    p += block



import struct
from construct import *

from m1n1.utils import irange
from m1n1.hw.dart import DART
from m1n1.utils import chexdump
from m1n1.proxyutils import RegMonitor
from m1n1.constructutils import *
from m1n1.hv.types import MMIOTraceFlags

from m1n1.trace.pcie import *

PCIeDevTracer = PCIeDevTracer._reloadcls()

mon = RegMonitor(hv.u)

class WLANCfgSpace(PCICfgSpace):
    PM_CSR              = 0x4c, Register32

    MSI_CAP             = 0x58, Register16
    MSI_CTRL            = 0x5a, Register16
    MSI_ADDR_L          = 0x5c, Register32
    MSI_ADDR_H          = 0x60, Register32
    MSI_DATA            = 0x64, Register32

    BAR0_WIN_1000       = 0x70, Register32
    BAR0_WIN_4000       = 0x74, Register32
    BAR0_WIN_5000       = 0x78, Register32
    BAR0_WINDOW         = 0x80, Register32
    BAR1_WINDOW         = 0x84, Register32
    SPROM_CONTROL       = 0x88, Register32
    CFG_SUBSYS_CONTROL  = 0x8c, Register32
    INTSTATUS           = 0x90, Register32
    INTMASK             = 0x94, Register32
    BACKPLANE_ADDR      = 0x98, Register32
    BACKPLANE_DATA      = 0x9c, Register32
    CLK_CTL_ST          = 0xa8, Register32

    CFG_DEVICE_CONTROL  = 0xb4, Register32
    LINK_STATUS_CTRL    = 0xbc, Register32

class WLANSRAMEnd(RegMap):
    PAD                 = 0x00, Register32
    SHARED_BASE         = 0x04, Register32

class WLANSRAMShared(RegMap):
    FLAGS               = 0, Register32
    CONSOLE_ADDR        = 20, Register32
    FWID                = 28, Register32
    MAX_RXBUFPOST       = 34, Register16
    RX_DATAOFFSET       = 36, Register32
    HTOD_MB_DATA_ADDR   = 40, Register32
    DTOH_MB_DATA_ADDR   = 44, Register32
    RING_INFO_ADDR      = 48, Register32
    DMA_SCRATCH_LEN     = 52, Register32
    DMA_SCRATCH_ADDR    = 56, Register64
    HOST_SCB_ADDR       = 64, Register64
    HOST_SCB_SIZE       = 72, Register32
    BUZZ_DBG_PTR        = 76, Register32
    FLAGS2              = 80, Register32
    HOST_CAP            = 84, Register32
    HOST_TRAP_ADDR      = 88, Register64
    DEVICE_FATAL_LOGBUF_START = 96, Register32
    HOFFLOAD_ADDR       = 100, Register64
    FLAGS3              = 108, Register32
    HOST_CAP2           = 112, Register32
    HOST_CAP3           = 116, Register32
    ETD_ADDR            = 120, Register32
    DEVICE_TXPOST_EXT_TAGS_BITMASK = 124, Register32

class WLANSRAMRingInfo(RegMap):
    RINGMEM             = 0x00, Register32
    H2D_W_IDX_PTR       = 0x04, Register32
    H2D_R_IDX_PTR       = 0x08, Register32
    D2H_W_IDX_PTR       = 0x0c, Register32
    D2H_R_IDX_PTR       = 0x10, Register32
    H2D_W_IDX_HOSTADDR  = 0x14, Register64
    H2D_R_IDX_HOSTADDR  = 0x1c, Register64
    D2H_W_IDX_HOSTADDR  = 0x24, Register64
    D2H_R_IDX_HOSTADDR  = 0x2c, Register64
    MAX_FLOWRINGS       = 0x34, Register32
    MAX_SUBMISSIONRINGS = 0x38, Register32
    MAX_COMPLETIONRINGS = 0x3c, Register32

COMMON_RING_CNT = 5

class WLANSRAMRingMem(RegMap):
    MAX_ITEM            = irange(0x04, COMMON_RING_CNT, 0x10), Register16
    LEN_ITEMS           = irange(0x06, COMMON_RING_CNT, 0x10), Register16
    BASE_ADDR           = irange(0x08, COMMON_RING_CNT, 0x10), Register32

class MsgHeader(ConstructClass):
    subcon = Struct(
        "msg_type" / Int8ul,
        "if_id" / Int8sl,
        "flags" / Int8ul,
        "epoch" / Int8ul,
        "request_id" / Int32ul,
    )

class ComplHeader(ConstructClass):
    subcon = Struct(
        "status" / Int16ul,
        "ring_id" / Int16ul,
    )

class IOCtlPtrReq(ConstructClass):
    subcon = Struct(
        "cmd" / Int32ul,
        "trans_id" / Int16ul,
        "input_buf_len" / Int16ul,
        "output_buf_len" / Int16ul,
        "rsvd" / Array(3, Int16ul),
        "host_input_buf_addr" / Int64ul,
    )

class IOCtlResp(ConstructClass):
    subcon = Struct(
        "compl" / ComplHeader,
        "resp_len" / Int16ul,
        "trans_id" / Int16ul,
        "cmd" / Int32ul,
    )

class H2DMailboxData(ConstructClass):
    subcon = Struct(
        "data" / Int32ul
    )

class D2HMailboxData(ConstructClass):
    subcon = Struct(
        "compl" / ComplHeader,
        "data" / Int32ul,
    )

class RingMessage(ConstructClass):
    subcon = Struct(
        "hdr" / MsgHeader,
        "payload" / Switch(this.hdr.msg_type, {
            0x09: IOCtlPtrReq,
            0x0c: IOCtlResp,
            0x23: H2DMailboxData,
            0x24: D2HMailboxData,

        }, default=HexDump(GreedyBytes))
    )

class RingState:
    pass

class WLANRingTracer(PCIeDevTracer):
    def __init__(self, wlan, info):
        self.wlan = wlan
        self.hv = wlan.hv
        self.p = wlan.hv.p
        self.u = wlan.hv.u
        self.ringid = self.RX, self.PTR_IDX
        if self.ringid in wlan.state.rings:
            self.state = wlan.state.rings[self.ringid]
        else:
            self.state = wlan.state.rings[self.ringid] = RingState()
            self.state.rptr = 0

        self.info = info
        assert info.item_size == self.ITEM_SIZE
        self.base_addr = info.base_addr
        self.count = info.count

        if self.RX:
            d2h_paddr = self.wlan.iotranslate(self.wlan.state.d2h_w_idx_ha + 4 * self.PTR_IDX, 4)[0][0]
            assert d2h_paddr is not None
            self.hv.add_tracer(irange(d2h_paddr, 4), self.wlan.ident, TraceMode.SYNC,
                               read=self.d2h_w_idx_readhook)

    def d2h_w_idx_readhook(self, evt):
        self.log("W idx read")
        self.poll()

    def poll(self):
        if self.RX:
            wptr = self.wlan.ioread(self.wlan.state.d2h_w_idx_ha + 4 * self.PTR_IDX, 4)
        else:
            wptr = self.wlan.ioread(self.wlan.state.h2d_w_idx_ha + 4 * self.PTR_IDX, 4)

        wptr = struct.unpack("<I", wptr)[0]

        while wptr != self.state.rptr:
            off = self.state.rptr * self.ITEM_SIZE
            addr = self.base_addr + off
            data = self.wlan.ioread(addr, self.ITEM_SIZE)
            self.pkt(data)
            self.state.rptr = (self.state.rptr + 1) % self.count

    def pkt(self, data):
        self.log("Got packet:")
        pkt = RingMessage.parse(data)
        self.log(pkt)
        if pkt.hdr.msg_type == 0x09:
            self.wlan.ioctlptr_req(pkt)
        if pkt.hdr.msg_type == 0x0c:
            self.wlan.ioctlresp(pkt)

    def log(self, msg):
        self.wlan.log(f"[{self.NAME}]{msg!s}")

class WLANControlSubmitRingTracer(WLANRingTracer):
    NAME = "CTLSubmit"
    PTR_IDX = 0
    RX = False
    ITEM_SIZE = 0x28

class WLANControlCompleteRingTracer(WLANRingTracer):
    NAME = "CTLCompl"
    PTR_IDX = 0
    RX = True
    ITEM_SIZE = 0x18

class RingInfo:
    def __init__(self):
        self.count = None
        self.item_size = None
        self.base_addr = None

    def ready(self):
        return self.count is not None and self.item_size is not None and self.base_addr is not None

class REG_IOCTL(Register32):
    CLK = 0
    FGC = 1
    CORE_SPECIFIC = 13, 2
    PME_EN = 14
    BIST_EN = 15

class REG_IOST(Register32):
    CORE_SPECIFIC = 11, 0
    DMA64 = 12
    GATED_CLK = 13
    BIST_ERROR = 14
    BIST_DONE = 15

class REG_RESET_CTL(Register32):
    RESET = 0

class REG_CLK_CTL(Register32):
    FORCEALP = 0
    FORCEHT = 1
    FORCEILP = 2
    HAVEALPREQ = 3
    HAVEHTREQ = 4
    HWCROFF = 5
    HQCLKREQ = 6
    EXTRESREQ = 11, 8
    HAVEALP = 16
    HAVEHT = 17
    BP_ON_ALP = 18
    BP_ON_HT = 19
    EXTRESST = 26, 24

class REG_PWR_CTL(Register32):
    DMN0 = 0
    DMN1 = 1
    DMN2 = 2
    DMN3 = 3
    DMN4 = 4
    PWRON_DMN0 = 8
    PWRON_DMN1 = 9
    PWRON_DMN2 = 10
    PWRON_DMN3 = 11
    PWRON_DMN4 = 12
    ST_DMN0 = 16
    ST_DMN1 = 17
    ST_DMN2 = 18
    ST_DMN3 = 19
    ST_DMN4 = 20
    BT_STATUS = 21, 20
    DMN_ID = 31, 28

class WLANAgentRegs(RegMap):
    IOCTL               = 0x408, REG_IOCTL
    IOST                = 0x500, REG_IOST
    RESET_CTL           = 0x800, REG_RESET_CTL

class WLANDevice(Reloadable):
    LENGTH = 0x1000
    WRAP_LENGTH = 0x1000
    REGMAP = None
    WRAP_REGMAP = WLANAgentRegs

    def __init__(self, wlan, type, rev, name, base, wrap):
        self.wlan = wlan
        self.hv = wlan.hv
        self.type = type
        self.rev = rev
        self.name = name
        self.base = base
        self.wrap = wrap
        self.regmap = None
        self.wrap_regmap = None

        if self.REGMAP is not None:
            self.regmap = self.REGMAP(self, base)

        if self.WRAP_REGMAP is not None:
            self.wrap_regmap = self.WRAP_REGMAP(self, wrap)

    def read(self, addr, width):
        assert False

    def write(self, addr, data, width):
        assert False

    def evt_rw(self, evt, base):
        reg = rcls = None
        addr = evt.addr + base
        value = evt.data

        t = "w" if evt.flags.WRITE else "r"

        pfx = f"{self.name}:"

        if self.regmap is not None:
            reg, index, rcls = self.regmap.lookup_addr(addr)
            regmap = self.regmap
            if rcls is not None:
                value = rcls(evt.data)

        if reg is None and self.wrap_regmap is not None:
            reg, index, rcls = self.wrap_regmap.lookup_addr(addr)
            regmap = self.wrap_regmap
            if rcls is not None:
                value = rcls(evt.data)
            if reg is not None:
                pfx = f"{self.name}.wrap:"

        if self.wlan.verbose >= 3 or (reg is None and self.wlan.verbose >= 1):
            if reg is None:
                s = pfx + f"{addr:#x} = {value:#x}"
            else:
                s = pfx + f"{regmap.get_name(addr)} = {value!s}"
            m = "+" if evt.flags.MULTI else " "
            self.wlan.log(f"WLAN: {t.upper()}.{1<<evt.flags.WIDTH:<2}{m} " + s)

        if reg is not None:
            attr = f"{t}_{reg}"
            handler = getattr(self, attr, None)
            if handler:
                if index is not None:
                    handler(value, index)
                else:
                    handler(value)
            elif self.wlan.verbose == 2:
                s = pfx + f"{regmap.get_name(evt.addr)} = {value!s}"
                m = "+" if evt.flags.MULTI else " "
                self.wlan.log(f"WLAN: {t.upper()}.{1<<evt.flags.WIDTH:<2}{m} " + s)

    def log(self, msg, show_cpu=True):
        self.wlan.log(f"[{self.name}] {msg}", show_cpu=show_cpu)

class WLANBackplane(Reloadable):
    SRAM_BASE = 0x740000
    SRAM_SIZE = 0x1f9000

    DEVICES = []

    def __init__(self, wlan):
        self.wlan = wlan
        self.hv = wlan.hv
        self.devices = []
        self.addr_map = ScalarRangeMap()
        self.name_map = {}
        self.unk_dev = WLANDevice(wlan, 0, 0, "unk", 0, 0)

        for type, rev, base, wrap, name, cls in self.DEVICES:
            if cls is None:
                cls = WLANDevice
            dev = cls(wlan, type, rev, name, base, wrap)
            self.devices.append(dev)
            self.wlan.log(f"add dev {name} base {base:#x} wrap {wrap:#x}")
            self.name_map[name] = dev
            if base != 0:
                self.addr_map[irange(base, dev.LENGTH)] = dev
            if wrap != 0:
                self.addr_map[irange(wrap, dev.LENGTH)] = dev
            setattr(self, name, dev)

    def evt_rw(self, evt, base):
        if isinstance(base, str):
            dev = self.name_map.get(base, self.unk_dev)
            base = dev.base
        else:
            dev = self.addr_map.get(base, self.unk_dev)

        dev.evt_rw(evt, base)
        regmap = dev.regmap

class WLANChipCommonRegs(RegMap):
    pass

class WLANPCIE2Regs(RegMap):
    INTMASK                 = 0x24, Register32
    MAILBOXINT              = 0x48, Register32
    MAILBOXMASK             = 0x4c, Register32
    CONFIGADDR              = 0x120, Register32
    CONFIGDATA              = 0x124, Register32
    H2D_MAILBOX_0           = 0x140, Register32
    H2D_MAILBOX_1           = 0x144, Register32

    CLK_CTL                 = 0x1e0, REG_CLK_CTL
    PWR_CTL                 = 0x1e8, REG_PWR_CTL

    HMAP_WINDOW_BASE_L      = 0x540, Register32
    HMAP_WINDOW_BASE_H      = 0x544, Register32
    HMAP_WINDOW_SIZE        = 0x548, Register32
    HMAP_VIOLATION_ADDR_L   = 0x5c0, Register32
    HMAP_VIOLATION_ADDR_H   = 0x5c4, Register32
    HMAP_VIOLATION_INFO     = 0x5c8, Register32
    HMAP_WINDOW_CONFIG      = 0x5d0, Register32

    HMAP_WINDOW_BASE_L_64   = 0x580, Register32
    HMAP_WINDOW_BASE_H_64   = 0x584, Register32
    HMAP_WINDOW_SIZE        = 0x588, Register32
    HMAP_VIOLATION_ADDR_L_64= 0x600, Register32
    HMAP_VIOLATION_ADDR_H_64= 0x604, Register32
    HMAP_VIOLATION_INFO_64  = 0x608, Register32
    HMAP_WINDOW_CONFIG_64   = 0x610, Register32

    H2D_MAILBOX_0_64        = 0xa20, Register32
    H2D_MAILBOX_1_64        = 0xa24, Register32
    INTMASK_64              = 0xc14, Register32
    MAILBOXINT_64           = 0xc30, Register32
    MAILBOXMASK_64          = 0xc34, Register32

class WLANPCIE2Core(WLANDevice):
    REGMAP = WLANPCIE2Regs

    def w_H2D_MAILBOX_0(self, val):
        ring = self.wlan.rings.get(2, None)
        if ring is not None:
            ring.poll()

        ring = self.wlan.rings.get(0, None)
        if ring is not None:
            ring.poll()

    w_H2D_MAILBOX_0_64 = w_H2D_MAILBOX_0

    def w_MAILBOXMASK_64(self, val):
        pass
        #self.hv.run_shell()

class WLANGCIRegs(RegMap):
    OTPDATA = irange(0x1000, 0x400, 2), Register16

class WLANGCICore(WLANDevice):
    LENGTH = 0x2000
    REGMAP = WLANGCIRegs

class WLANBackplane4388(WLANBackplane):
    SRAM_BASE = 0x200000
    SRAM_SIZE = 0x2e0000

    DEVICES = [
        (0x800, 75,  0x18000000, 0x18100000, "cc",      None),
        (0x83c, 74,  0x18001000, 0x18101000, "pcie2",   WLANPCIE2Core),
        (0x847, 11,  0x18020000, 0x18120000, "arm_ca7", None),
        (0x812, 87,  0x18021000, 0x18121000, "80211_0", None),
        (0x812, 87,  0x18022000, 0x18122000, "80211_1", None),
        (0x812, 87,  0x18023000, 0x18123000, "80211_2", None),
        (0x850, 1,   0x00000000, 0x1812c000, "850?",    None),
        (0x849, 12,  0x18024000, 0x18124000, "sys_mem", None),
        (0x857, 8,   0x00000000, 0x18108000, "857?",    None),
        (0x840, 27,  0x18010000, 0x00000000, "gci",     WLANGCICore),
        (0x827, 43,  0x18012000, 0x00000000, "pmu",     None),
        (0x135, 0,   0x00000000, 0x18106000, "apb_0",  None),
        (0x135, 0,   0x00000000, 0x18107000, "apb_1",  None),
        (0x135, 0,   0x00000000, 0x18132000, "apb_2",  None),
        (0x31,  0,   0x00000000, 0x18131000, "31_0?",   None),
        (0x31,  0,   0x00000000, 0x1810b000, "31_1?",   None),
        (0x31,  0,   0x00000000, 0x1810c000, "31_2?",   None),
        (0xfff, 0,   0x1801f000, 0x18104000, "fff_0?",  None),
        (0xfff, 0,   0x1801f000, 0x1812f000, "fff_1?",  None),
    ]

class WLANTracer(PCIeDevTracer):
    DEFAULT_MODE = TraceMode.SYNC

    BARMAPS = [None, None, None]
    CFGMAP = WLANCfgSpace

    RINGS = [
        WLANControlSubmitRingTracer,
        None, # RXPost
        WLANControlCompleteRingTracer,
        None, # TX complete
        None, # RX complete
    ]

    CMDS = {
        1: "GET_VERSION",
        2: "UP",
        3: "DOWN",
        262: "GET_VAR",
        263: "SET_VAR",
    }

    CHIPSETS = {
        4388: WLANBackplane4388,
    }

    def __init__(self, hv, apcie, bus, dev, fn, dart_path=None, verbose=False):
        super().__init__(hv, apcie, bus, dev, fn, verbose=verbose)
        self.u = hv.u
        self.p = hv.p
        self.dart_path = dart_path
        self.dart_dev = None
        self.dart = None
        self.rings = {}
        self.chipset = u.adt["/product"].wifi_chipset
        self.bp = self.CHIPSETS.get(int(self.chipset), WLANBackplane)(self)

    def init_state(self):
        super().init_state()
        self.state.shared_base = None
        self.state.ring_info_base = None
        self.state.ring_mem_base = None
        self.state.tcm_base = None
        self.state.tcm_size = None
        self.state.ring_info = None
        self.state.ring_mem = None
        self.state.ring_info_data = {}
        self.state.rings = {}
        self.state.ioctls = {}
        self.state.h2d_w_idx_ha = None
        self.state.h2d_r_idx_ha = None
        self.state.d2h_w_idx_ha = None
        self.state.d2h_r_idx_ha = None

    def bar0_rw(self, evt):
        off = evt.addr - self.state.bar0_base
        value = evt.data

        t = "w" if evt.flags.WRITE else "r"

        evt = evt.copy()
        evt.addr = off & 0xfff

        if 0x0000 <= off < 0x1000:
            self.bp.evt_rw(evt, self.cfg.cached.BAR0_WINDOW.val)
        elif 0x1000 <= off < 0x2000:
            self.bp.evt_rw(evt, self.cfg.cached.BAR0_WIN_1000.val)
        elif 0x2000 <= off < 0x3000:
            self.bp.evt_rw(evt, "pcie2")
        elif 0x3000 <= off < 0x4000:
            self.bp.evt_rw(evt, "cc")
        elif 0x4000 <= off < 0x5000:
            self.bp.evt_rw(evt, self.cfg.cached.BAR0_WIN_4000.val)
        elif 0x5000 <= off < 0x6000:
            self.bp.evt_rw(evt, self.cfg.cached.BAR0_WIN_5000.val)
        else:
            self.bp.evt_rw(evt, f"unk{off>>12:#x}")

    def config_dart(self):
        # Ugly...
        if self.dart_dev is None:
            for i in range (16):
                ttbr = self.dart.dart.regs.TTBR[i].reg
                if ttbr.VALID:
                    self.log(f"DART device: {i}")
                    self.dart_dev = i
                    break
            else:
                raise Exception("Failed to find DART device")

    def ioread(self, addr, size):
        self.config_dart()
        return self.dart.ioread(self.dart_dev, addr, size)

    def iotranslate(self, addr, size):
        self.config_dart()
        return self.dart.iotranslate(self.dart_dev, addr, size)

    def r_SHARED_BASE(self, base):
        if base.value & 0xffff == (base.value >> 16) ^ 0xffff:
            return

        self.state.shared_base = base.value
        self.update_shared()

    def w_H2D_W_IDX_HOSTADDR(self, addr):
        self.state.h2d_w_idx_ha = addr.value

    def w_H2D_R_IDX_HOSTADDR(self, addr):
        self.state.h2d_r_idx_ha = addr.value

    def w_D2H_W_IDX_HOSTADDR(self, addr):
        self.state.d2h_w_idx_ha = addr.value

    def w_D2H_R_IDX_HOSTADDR(self, addr):
        self.state.d2h_r_idx_ha = addr.value

    def w_MAX_ITEM(self, val, index):
        info = self.state.ring_info_data.setdefault(index, RingInfo())
        info.count = val.value
        self.update_ring(index)

    def w_LEN_ITEMS(self, val, index):
        info = self.state.ring_info_data.setdefault(index, RingInfo())
        info.item_size = val.value
        self.update_ring(index)

    def w_BASE_ADDR(self, val, index):
        info = self.state.ring_info_data.setdefault(index, RingInfo())
        info.base_addr = val.value
        self.update_ring(index)

    def w_cfg_BACKPLANE_DATA(self, val):
        win = self.cfg.cached.BAR0_WINDOW.val
        evt = Container()
        evt.flags = MMIOTraceFlags(WIDTH=2, WRITE=1)
        evt.addr = self.cfg.cached.BACKPLANE_ADDR.val
        evt.data = val.value
        self.bp.evt_rw(evt, win)

    def r_cfg_BACKPLANE_DATA(self, val):
        win = self.cfg.cached.BAR0_WINDOW.val
        evt = Container()
        evt.flags = MMIOTraceFlags(WIDTH=2, WRITE=0)
        evt.addr = self.cfg.cached.BACKPLANE_ADDR.val
        evt.data = val.value
        self.bp.evt_rw(evt, win)

    def update_ring(self, idx):
        if idx not in self.state.ring_info_data:
            return
        info = self.state.ring_info_data[idx]
        if not info.ready():
            return

        if idx in self.rings:
            return

        if idx > len(self.RINGS):
            return

        ringcls = self.RINGS[idx]

        if ringcls is None:
            return

        self.rings[idx] = ringcls(self, info)

    def ioctlptr_req(self, pkt):
        data = self.ioread(pkt.payload.host_input_buf_addr, pkt.payload.input_buf_len)
        cmd = self.CMDS.get(pkt.payload.cmd, "unk")
        self.log(f"IOCTL request ({cmd}):")
        chexdump(data, print_fn = self.log)
        self.state.ioctls[pkt.payload.trans_id] = pkt

    def ioctlresp(self, pkt):
        req = self.state.ioctls.get(pkt.payload.trans_id, None)
        if req is None:
            self.log(f"ERROR: unknown transaction ID {pkt.payload.trans_id:#x}")
            return

        data = self.ioread(req.payload.host_input_buf_addr, req.payload.output_buf_len)
        cmd = self.CMDS.get(pkt.payload.cmd, "unk")
        self.log(f"IOCTL response ({cmd}):")
        chexdump(data, print_fn = self.log)
        del self.state.ioctls[pkt.payload.trans_id]

    def trace_bar(self, idx, start, size):
        if idx == 2:
            self.state.tcm_base = start
            self.state.tcm_size = size

            self.update_tcm_tracers()
        elif idx == 0:
            self.state.bar0_base = start
            self.state.bar0_size = size

            self.update_bar0_tracers()
        else:
            return super().trace_bar(idx, start, size)

    def update_tcm_tracers(self):
        if self.state.tcm_base is None:
            return

        if self.dart is None:
            self.dart = DART.from_adt(self.u, self.dart_path)

        self.trace_regmap(self.state.tcm_base + self.bp.SRAM_BASE + self.bp.SRAM_SIZE - 8, 8,
                          WLANSRAMEnd, name="sram")

    def update_bar0_tracers(self):
        if self.state.bar0_base is None:
            return

        zone = irange(self.state.bar0_base, self.state.bar0_size)
        self.hv.add_tracer(zone, self.ident, self.DEFAULT_MODE, self.bar0_rw,
                           self.bar0_rw)

    def update_shared(self):
        base = self.state.shared_base
        if base is None:
            return

        if self.state.ring_info_base is None:
            self.shared = WLANSRAMShared(self.hv.u, self.state.tcm_base + base)

            self.log("Reading shared info")
            self.shared.dump_regs()

            self.state.ring_info_base = self.shared.RING_INFO_ADDR.val

        if self.state.ring_mem_base is None:
            self.ring_info = WLANSRAMRingInfo(self.hv.u,
                                            self.state.tcm_base + self.state.ring_info_base)
            self.log("Reading ring info")
            self.ring_info.dump_regs()

            self.state.ring_mem_base = self.ring_info.RINGMEM.val

        self.trace_regmap(self.state.tcm_base + base, 0x100,
                          WLANSRAMShared, name="shared")

        self.trace_regmap(self.state.tcm_base + self.state.ring_info_base, 0x40,
                          WLANSRAMRingInfo, name="ringinfo")

        self.ring_mem = WLANSRAMRingMem(self.hv.u,
                                        self.state.tcm_base + self.state.ring_mem_base)
        self.log("Reading ring mem")
        self.ring_mem.dump_regs()

        self.trace_regmap(self.state.tcm_base + self.state.ring_mem_base,
                          COMMON_RING_CNT * 0x10, WLANSRAMRingMem, name="ringmem")

    def start(self):
        super().start()

        self.update_tcm_tracers()
        self.update_shared()
        for i in range(len(self.RINGS)):
            self.update_ring(i)

devno = 2 if hv.xnu_mode else 1

wlan_tracer = WLANTracer(hv, "/arm-io/apcie",
                         devno, 0, 0, "/arm-io/dart-apcie0")

wlan_tracer.start()



# SPDX-License-Identifier: MIT

from ..hv import TraceMode
from ..utils import *
from . import ADTDevTracer

class R_PIN(Register32):
    DRIVE_STRENGTH1 = 23, 22
    LOCK            = 21
    GROUP           = 18, 16
    SCHMITT         = 15
    DRIVE_STRENGTH0 = 11, 10
    INPUT_ENABLE    = 9
    PULL            = 8, 7
    PERIPH          = 6, 5
    MODE            = 3, 1
    DATA            = 0

class GPIORegs(RegMap):
    PIN = irange(0x000, 212, 4), R_PIN

    IRQ_GROUP  = (irange(0x800, 7, 0x40), irange(0, (212 + 31) // 32, 4)), Register32

def bits32(val, start):
    return [start + i for i in range(0, 32) if int(val) & (1 << i)]

class GPIOTracer(ADTDevTracer):
    DEFAULT_MODE = TraceMode.UNBUF

    REGMAPS = [GPIORegs]
    NAMES = ["gpio"]

    PIN_NAMES = {}

    def __init__(self, hv, devpath, pin_names={}, verbose=False):
        super().__init__(hv, devpath, verbose)
        self.PIN_NAMES = pin_names

    def pn(self, pin):
        return self.PIN_NAMES.get(pin, f"Pin-{pin}")

    def r_PIN(self, val, index):
        if index not in self.PIN_NAMES and self.verbose < 2:
            return
        self.log(f"{self.pn(index):14} R {val!s} ")

    def w_PIN(self, val, index):
        if index not in self.PIN_NAMES and self.verbose < 2:
            return
        self.log(f"{self.pn(index):14} W {val!s} ")

    def r_IRQ_GROUP(self, val, index):
        (grp, index) = index
        if int(val) == 0:
            return
        pins = [self.pn(x) for x in bits32(val, index * 32) if self.verbose >= 2 or x in self.PIN_NAMES]
        if len(pins):
            self.log(f"IRQ[{grp}] ACT {pins}")

    def w_IRQ_GROUP(self, val, index):
        (grp, index) = index
        if int(val) == 0:
            return
        pins = [self.pn(x) for x in bits32(val, index * 32) if self.verbose >= 2 or x in self.PIN_NAMES]
        if len(pins):
            self.log(f"IRQ[{grp}] ACK {pins}")




# SPDX-License-Identifier: MIT

from m1n1.trace.gpio import GPIOTracer

#trace_device("/arm-io/gpio", True)

# trace gpio interrupts, useful to follow the cascaded interrupts
aic_phandle = getattr(hv.adt["/arm-io/aic"], "AAPL,phandle")
try:
    node = hv.adt["/arm-io/gpio0"]
    path = "/arm-io/gpio0"
except:
    node = hv.adt["/arm-io/gpio"]
    path = "/arm-io/gpio"

if getattr(node, "interrupt-parent") == aic_phandle:
    for irq in getattr(node, "interrupts"):
        hv.trace_irq(node.name, irq, 1, hv.IRQTRACE_IRQ)

PIN_NAMES_j274 = {
    0xC0: "i2c0:scl",
    0xBC: "i2c0:sda",
    0xC9: "i2c1:scl",
    0xC7: "i2c1:sda",
    0xA3: "i2c2:scl",
    0xA2: "i2c2:sda",
    106:  "hpm:irq",
    136:  "bluetooth:irq",
    196:  "wlan:irq",
    183:  "cs42l83:irq",
    182:  "tas5770:irq",
    152:  "pci@0,0",
    153:  "pci@1,0",
    33:   "pci@2,0",
    #0x2D: "spi_nor:CS",
}

GPIOTracer = GPIOTracer._reloadcls()
gpio_tracer = GPIOTracer(hv, path, PIN_NAMES_j274, verbose=0)
gpio_tracer.start()
